class Polynomial2():
    '''
    模二多项式环，定义方式有三种
    一是
        >>> Polynomial2([1,1,0,1]) # 指定每一项的系数，不能为空
        x^3 + x^2 + 1

    二是
        >>> Polynomial2('1101')
        x^3 + x^2 + 1

    三是
        >>> Poly([3,1,4]) # 直接给出系数为1的项的阶
        x^4 + x^3 + x
        >>> Poly()        # 可以为空，代表所有项系数为0，也就是0
        0
        >>> Poly(9,6,5)
        x^9 + x^6 + x^5
    '''
    def __init__(self,ll):
        
        if type(ll) ==  str:
            ll = list(map(int,ll))

        self.param = ll[::-1]
        self.ones = [i for i in range(len(self.param)) if self.param[i] == 1] # 系数为1的项的阶数列表
        self.Latex = self.latex()
        self.b = ''.join([str(i) for i in ll]) # 二进制形式打印系数
        
        self.order = 0 # 最高阶
        try:self.order = max(self.ones)
        except:pass
        
    def format(self,reverse = True):
        '''
            格式化打印字符串
            
            reverse = False时，可以低位在左
            但是注意定义多项式时只能高位在右
        '''
        r = ''
        if len(self.ones) == 0:
            return '0'
        if reverse:
            return ((' + '.join(f'x^{i}' for i in self.ones[::-1])+' ').replace('x^0','1').replace('x^1 ','x ')).strip()
        return ((' + '.join(f'x^{i}' for i in self.ones)+' ').replace('x^0','1').replace('x^1 ','x ')).strip()

    def __call__(self,x):# 懒得写
        '''
            懒得写了
        '''
        print(f'call({x})')

    def __add__(self,other):
        a,b = self.param[::-1],other.param[::-1]
        if len(a) < len(b):a,b = b,a
        for i in range(len(a)):
            try:a[-1-i] = (b[-1-i] + a[-1-i]) % 2
            except:break
        return Polynomial2(a)

    def __mul__(self,other):
        a,b = self.param[::-1],other.param[::-1]
        r = [0 for i in range(len(a) + len(b) - 1)]
        for i in range(len(b)):
            if b[-i-1] == 1:
                if i != 0:sa = a+[0]*i
                else:sa = a
                sa = [0] * (len(r)-len(sa)) + sa
                #r += np.array(sa)
                #r %= 2
                r = [(r[t] + sa[t])%2 for t in range(len(r))]
        return Polynomial2(r)

    def __sub__(self,oo):
        # 在模二多项式环下，加减相同
        return self + oo

    def div(self,other):
        r,b = self.param[::-1],other.param[::-1]
        if len(r) < len(b):
            return Polynomial2([0]),self

        q=[0] * (len(r) - len(b) + 1)
        for i in range(len(q)):
            if len(r)>=len(b):
                index = len(r) - len(b) + 1  # 确定所得商是商式的第index位
                q[-index] = int(r[0] / b[0])
                # 更新被除多项式
                b_=b.copy()
                b_.extend([0] * (len(r) - len(b)))
                b_ = [t*q[i] for t in b_] 
                r = [(r[t] - b_[t])%2 for t in range(len(r))]
                for j in range(len(r)):     #除去列表最左端无意义的0
                    if r[0]==0:
                        r.remove(0)
                    else:
                        break
            else:
                break
            
        return Polynomial2(q),Polynomial2(r)

    def __floordiv__(self,other): # 只重载了整除，即//
        return self.div(other)[0]

    def __mod__(self,other):
        return self.div(other)[1]

    def __repr__(self) -> str:
        return self.format()
    
    def __str__(self) -> str:
        return self.format()

    def __pow__(self,a):
        # 没有大数阶乘的需求，就没写快速幂
        t = Polynomial2([1])
        for i in range(a):
            t *= self
        return t
    
    def latex(self,reverse=True):
        '''
            Latex格式打印...其实就是给大于一位长度的数字加个括号{}
        '''
        def latex_pow(x):
            if len(str(x)) <= 1:
                return str(x)
            return '{'+str(x)+'}'
        
        r = ''
        if len(self.ones) == 0:
            return '0'
        if reverse:
            return (' + '.join(f'x^{latex_pow(i)}' for i in self.ones[::-1])+' ').replace('x^0','1').replace(' x^1 ',' x ').strip()
        return (' + '.join(f'x^{latex_pow(i)}' for i in self.ones)+' ').replace('x^0','1').replace(' x^1 ',' x ').strip()

    def __eq__(self,other):
        return self.ones == other.ones


def Poly(*args):
    '''
        另一种定义方式
        Poly([3,1,4]) 或 Poly(3,1,4)
    '''
    if len(args) == 1 and type(args[0]) in [list,tuple]:
        args = args[0]
        
    if len(args) == 0:
        return Polynomial2('0')
    ll = [0 for i in range(max(args)+1)]
    for i in args:
        ll[i] = 1
    return Polynomial2(ll[::-1])

PP = Polynomial2

P = Poly



from tqdm import *
from time import sleep
import time
from random import randint,choice
from os import system,urandom
from termcolor import colored
import f61d.LOGO as LOGO
#from f61d.__version__ import *
from math import ceil
from tqdm import *

VERSION = (0,3,0)
VERSION = '.'.join(map(str, VERSION))



print(colored(LOGO.randlogo()+' v'+VERSION+'\n\n',choice(['red','yellow','green'])))
print(colored("Welcome to F61D. Use Help() for help.",'green'))

help_cont = {}
help_cont[0] = 'Help     : See this help page.\n\t\t  '+colored('Help(0)','cyan')+colored(' for all functions.','green')
help_cont[1] = 'fuckStar : One Button to fuck Satellite'
help_cont[9] = 'lifeGame : Play the Life Game'
help_cont[3] = 'paint    : Draw a Bing Dwen Dwen'
help_cont[4] = 'b64tofile: Decode base64 to a file'
help_cont[5] = 'prove    : Prove your self. Only support:\n\t\t  sha256(XXXX + abcdef) = aaabbb...'
help_cont[6] = 'Nim      : An interesting turtle game.'
help_cont[7] = 'getLogo  : Print the "F61D" LOGO.'
help_cont[8] = 'yinyang  : Draw the TaiJi.'
help_cont[2] = 'timeit   : Take a time for your function.\n\t\t  Just add a '+colored('@timeit','cyan')+colored(' before you define the function','green')
help_cont[10] = 'Polynomial: from f61d.Polynomial2 import *'


def Help(page=1,lines = 5):
    if type(page) != int or type(lines) != int:
        Help(1,10000)
        return print('Please input interger.')
    if page == 0:
        return Help(1,10000)
    maxPage = ceil(len(help_cont)/lines)#int((0.5+len(help_cont))/lines)+1
    if (page - 1) * lines >= len(help_cont):
        print(f'Page Error, check the number : [1,{maxPage}]')
        return
    print(colored(f"{'=-'*14}= ",'yellow')+"H E L P"+colored(f" {'=-'*14}=\n",'yellow'))
    for i in range(lines):
        try:
            n = i+lines*(page-1)
            ll = colored(help_cont[n],'green')
            print(f'   {n} - '+ll)
        except:break
    print('\n'+colored(f"{'=-'*14}= ",'yellow')+f"page({page}/{maxPage})"+colored(f" {'=-'*14}=",'yellow'))
    if page < maxPage:print(colored(f'\t\t\thelp({page+1}) for next page','green'))


def fuckStar():
    '''
一键日卫星
'''
    print(' - 欢迎使用一键日卫星功能')
    print(' - 请输入要日的卫星编号(可在官网查询):',end='')
    num = input()
    while num == '':
        num = input('请重新输入：')

    bar = trange(randint(1000,2000))
    for i in bar:
        bar.set_description('Hacking:'+hex(int.from_bytes(urandom(4),'big'))[2:])
        for _ in range(66666):
            a = 1
            b = 1
            a,b = b,1
    for i in range(3):
        print('.',end='')
        sleep(0.4)
    if randint(1,100) > 80 or num == 'CHY':
        for i in range(3):
            print(colored('.','green'),end='')
            sleep(0.4)
        print(colored('\n日卫星成功.','green'))    
    else:
        print(colored('\n日卫星失败，请重新尝试','red'))
    
def lifeGame():
    from f61d.lifeGame import game
    game()
    
def paint():
    from f61d.bingdwendwen import draw
    draw()
    
def b64tofile():
    '''
    Encode/Decode a file by Base64
    '''
    menu = '''
please choose mode (1/2):
    - 1. encode
        Base64 encode a file
    - 2. decode
        Base64 decode a file
'''
    mode = input(menu)
    f1 = input('input file name:')
    f2 = input('Output file name:')
    while mode not in ['1','2']:
        mode = input('Please choose mode (1/2):')
    f = open(f1,'rb').read()
    from base64 import b64decode,b64encode
    if mode == '1':
        df = b64decode(f)
    elif mode == '2':
        df = b64encode(f)
    with open(f2,'wb') as ff:
        ff.write(df)


def Nim():
    from f61d.nim import game
    game()


def timeit(fun):
    '''
计算你的函数运行时间
'''
    def wrapper(*args,**kwargs):
        
        t1 = time.time()
        ret = fun(*args,**kwargs)
        print(f'Function takes {time.time() - t1}s')

        return ret
    
    return wrapper


def prove(cont,process_bar=True,show_description=False):
    '''
    Only the format is supported
    sha256(XXXX + {a}) = {b}
    '''
    if process_bar and show_description:
        print("Open tqdm bar and description may be VERY SLOW.\nDescription is not sugested.")
    from itertools import product
    import re
    from string import ascii_letters,digits
    from hashlib import sha256

    
    cont = cont.replace(' ','')
    #print(cont)
    p = re.compile('sha256(.*?)==(.*)')
    #print(p.findall(cont)[0])
    try:
        r = p.findall(cont)[0]
    except:
        print('Failed to parse text.')
        return
    L = len(r[0].split('+')[0])-1
    a = r[0].split('+')[1][:-1]
    b = r[1]
    print(f"Proving:\n\tsha256({'X'*L} + {a}) = {b}\n")
    s = list(product(ascii_letters+digits,repeat=L))
    if process_bar:bar = tqdm(s)
    else:bar = s
    ss = lambda d:sha256(d.encode()).hexdigest()
    for i in bar:
        S = ''.join(i)
        if process_bar and show_description:bar.set_description(f"trying: {S}")
        if ss(S+a) == b:
            print('\nFind ,',S)
            return S
    else:
        print('Failed to find \'XXXX\'')
        return None

    
def getLogo():
    print(colored(LOGO.randlogo(),choice(['red','yellow','green'])))
    #print(__verion__)

def yinyang():
    import f61d.yinyang
    f61d.yinyang.main()
    
    
